/* ===============================================================================
     File: @(#)ascos_public.cpp
     Author: Christoph Hormann <chris_hormann@gmx.de>
     Creation Date: $ 31 Mar 2006, 13:53:58 $
     Last modified: $ 08 Nov 2012, 14:55:28 $
   -------------------------------------------------------------------------------
     Description: atmosphere and shading correction 
                  of satellite-images

		 Version history:

		   0.1: initial version
		   0.2: adaptation to CImg 1.14
		   0.3: adaptation to CImg 1.21 and GPLv3 license change
		   0.4: adaptation to CImg 1.29
		   0.5: adaptation to CImg 1.5

   -------------------------------------------------------------------------------

     Copyright (C) 2006-2012 Christoph Hormann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.

   ===============================================================================
 */

#include <cstdlib>
#include <fstream>
#include <iostream>
#include <algorithm>

#define cimg_use_tiff 1

#include <CImg.h>

using namespace cimg_library;

const char PROGRAM_TITLE[] = "Atmosphere and shading correction of satellite-images version 0.5";

const unsigned char col_bkg[3] = {80,80,80};
const unsigned char col_text[3] = {200,200,200};
const unsigned char col_text1[3] = {255,255,255};
const unsigned char col_text2[3] = {255,220,200};
const unsigned char col_text3[3] = {210,210,255};

const float GAMMA = 2.2;

int main(int argc,char **argv)
{
	std::fprintf(stderr,"------------------------------------------------------------------\n");
	std::fprintf(stderr,"%s\n", PROGRAM_TITLE);
	std::fprintf(stderr,"------------------------------------------------------------------\n");

	cimg_usage("Usage : ascos [options]");

	// --- Read command line parameters ---
	const bool Interactive_Mode = cimg_option("-interactive",false,"Interactive mode");
	float Brightness = cimg_option("-b",1.0,"Display brightness adjustment");

	// Files
	const char *file_i = cimg_option("-i",(char*)NULL,"Color Image");
	const char *file_dem = cimg_option("-d",(char*)NULL,"DEM Image");
	const char *file_o = cimg_option("-o",(char*)NULL,"Output Image");

	const bool do_16 = cimg_option("-16",false,"Generate 16 bit output");

	// Color Adjustments
	float Gamma = cimg_option("-gamma",GAMMA,"Gamma correction");
	bool do_gamma = true;
	if (Gamma < 0.0)
	{
		Gamma = -Gamma;
		do_gamma = false;
	}

	// Atmosphere correction
	float Weights[3];
	const char *wgt_string = cimg_option("-w","0.1:0.1:0.1","Correction weights");
	std::sscanf(wgt_string,"%f:%f:%f",&Weights[0],&Weights[1],&Weights[2]);
	int HMax = cimg_option("-r",-1,"Reference height");
	const int HZero = cimg_option("-z",32767,"Zero height");
	const float HMult = cimg_option("-m",1.0,"Height multiplier");
	float Exponent = cimg_option("-e",1.0,"Correction Exponent");

	// Shading correction
	const float GridSz_Col = cimg_option("-resolution",30.0,"Resolution of color image (same unit as DEM values)");

	const float LightAzimuth = cimg_option("-azimuth",0.0,"Sun azimuth");
	const float LightElevation = cimg_option("-elevation",90.0,"Sun elevation");

	float LightAdjust[3];
	const char *lgt_string = cimg_option("-l","0.1:0.1:0.1","Lighting compensation weights");
	std::sscanf(lgt_string,"%f:%f:%f",&LightAdjust[0],&LightAdjust[1],&LightAdjust[2]);

	const bool helpflag = cimg_option("-h",false,"Display this help");
	if (helpflag) std::exit(0);

	if ((file_i == NULL) || (file_dem == NULL))
	{
		std::fprintf(stderr,"You must specify a color and dem image (try '%s -h').\n\n",argv[0]);
		std::exit(0); 
	}

	// --- Load input images ---
	std::fprintf(stderr,"Loading color image...\n");
	CImg<unsigned char> img_col = CImg<unsigned char>(file_i);
	std::fprintf(stderr,"Loading DEM image...\n");
	CImg<unsigned short> img_dem = CImg<unsigned short>(file_dem);
	CImg<> img_proc;
	if (do_16) img_proc = CImg<unsigned short>(img_col);
	else img_proc = CImg<unsigned char>(img_col);

	if (HMax < 0)
	{
		cimg_forY(img_dem,py)
		{
			cimg_forX(img_dem,px)
			{
				if (img_dem(px,py)>HMax) HMax = img_dem(px,py);
			}
		}
		HMax = int(HMax/HMult) - HZero;
	}

	const float GridSz_DEM = GridSz_Col*img_col.width()/img_dem.width();

	float LightDir[3];
	LightDir[0] = sin((90.0-LightElevation)*cimg::PI/180.0)*sin(LightAzimuth*cimg::PI/180.0);
	LightDir[1] = sin((90.0-LightElevation)*cimg::PI/180.0)*cos(LightAzimuth*cimg::PI/180.0);
	LightDir[2] = cos((90.0-LightElevation)*cimg::PI/180.0);

	std::fprintf(stderr,"\n");
	std::fprintf(stderr,"  Weights:              <%.2f, %.2f, %.2f>\n",Weights[0], Weights[1], Weights[2]);
	std::fprintf(stderr,"  Correction Exponent:  %.2f\n",Exponent);
	std::fprintf(stderr,"  Light Compensation:   <%.2f, %.2f, %.2f>\n",LightAdjust[0], LightAdjust[1], LightAdjust[2]);
	std::fprintf(stderr,"  Reference Height:     %d\n",HMax);
	std::fprintf(stderr,"  Display Brightness:   %.2f\n",Brightness);
	std::fprintf(stderr,"\n");
	std::fprintf(stderr,"  Light Dir:            <%.2f, %.2f, %.2f>\n",LightDir[0], LightDir[1], LightDir[2]);
	std::fprintf(stderr,"  Color Img Resolution: %.2f\n", GridSz_Col);
	std::fprintf(stderr,"  DEM Resolution:       %.2f\n", GridSz_DEM);
	std::fprintf(stderr,"  Zero Height:          %d\n", HZero);
	std::fprintf(stderr,"  Height Multiplier:    %.2f\n", HMult);
	std::fprintf(stderr,"\n");

	bool Save = true;

	if (Interactive_Mode)
	{
		bool stopflag = false;
		bool line_redraw = true;
		bool redraw = true;
		int Mx = 0, My = 0;
		int Zx0 = 0, Zy0 = 0, Zx1, Zy1;
		int Zx0d, Zy0d, Zx1d, Zy1d;
		int Adjust = 2;
		bool Adjust2 = false;
		bool disp_dem = false;
		bool zoom = false;

		int dispsz_x, dispsz_y;
		bool img_scaled = false;
		CImg<unsigned char> img_col_s;
		CImg<unsigned short> img_dem_s;

		if ((img_col.width() > CImgDisplay::screen_width()) || (img_col.height() > CImgDisplay::screen_height()))
		{
			dispsz_x = CImgDisplay::screen_width()-400;
			dispsz_y = dispsz_x*img_col.height()/img_col.width();
			img_col_s = img_col.get_resize(dispsz_x, dispsz_y);
			img_dem_s = img_dem.get_resize(dispsz_x, dispsz_y);
			img_scaled = true;
		}
		else
		{
			dispsz_x = img_col.width();
			dispsz_y = img_col.height();
			img_col_s = img_col;
			img_dem_s = img_dem;
		}

		CImg<unsigned char> img_disp = CImg<unsigned char>(img_col_s);
		CImg<unsigned char> img_proc_s = CImg<unsigned char>(img_col_s);
		CImgDisplay disp(img_disp,PROGRAM_TITLE);

		while(!stopflag)
		{
		switch (disp.key())
		{
		  case cimg::keySPACE:
				disp_dem = !disp_dem;
				redraw = true;
				break;
		  case cimg::keyZ:
				if ((disp.mouse_x() >= 0) && img_scaled)
				{
					zoom = !zoom;
					if (zoom)
					{
						float Zx = (float)disp.mouse_x()/dispsz_x;
						float Zy = (float)disp.mouse_y()/dispsz_y;
						Zx0 = (int)(Zx*img_proc.width()-dispsz_x/2);
						Zy0 = (int)(Zy*img_proc.height()-dispsz_x/2);
						Zx1 = Zx0+dispsz_x;
						Zy1 = Zy0+dispsz_y;
						Zx0d = Zx0*img_dem.width()/img_proc.width();
						Zy0d = Zy0*img_dem.height()/img_proc.height();
						Zx1d = Zx1*img_dem.width()/img_proc.width();
						Zy1d = Zy1*img_dem.height()/img_proc.height();
						img_col_s = img_col.get_crop(Zx0, Zy0, Zx1, Zy1);
						img_dem_s = img_dem.get_crop(Zx0d, Zy0d, Zx1d, Zy1d);
					}
					else
					{
						img_col_s = img_col.get_resize(dispsz_x, dispsz_y);
						img_dem_s = img_dem.get_resize(dispsz_x, dispsz_y);
					}
					img_proc_s = img_col_s;
					redraw = true;
				}
				break;
		  case cimg::keyX:
			{
				img_proc_s = img_col_s;
				redraw = true;
				break;
			}
		  case cimg::keyA:
			{
				float col[3];
				float Correction;

				bool doLight = true;
				if ((LightAdjust[0] < 0.01) && (LightAdjust[1] < 0.01) && (LightAdjust[2] < 0.01))
					doLight = false;

				cimg_forY(img_col_s,py)
				{
					float pyd = (float)py*img_dem_s.height()/img_col_s.height();
					cimg_forX(img_col_s,px)
					{
						float pxd = (float)px*img_dem_s.width()/img_col_s.width();

						float Height = img_dem_s.linear_atXY(pxd,pyd)/HMult-HZero;
						float AOT = ((float)HMax - Height)/float(HMax);

						col[0] = pow((float)img_col_s(px,py,0,0)/255.0, Gamma);
						col[1] = pow((float)img_col_s(px,py,0,1)/255.0, Gamma);
						col[2] = pow((float)img_col_s(px,py,0,2)/255.0, Gamma);

						col[0] -= pow(col[0], Exponent)*Weights[0]*AOT;
						col[1] -= pow(col[1], Exponent)*Weights[1]*AOT;
						col[2] -= pow(col[2], Exponent)*Weights[2]*AOT;

						if (doLight)
						{
							float Hgt_xm = img_dem_s.linear_atXY(pxd+0.1,pyd);
							float Hgt_xp = img_dem_s.linear_atXY(pxd-0.1,pyd);

							float Hgt_ym = img_dem_s.linear_atXY(pxd,pyd-0.1);
							float Hgt_yp = img_dem_s.linear_atXY(pxd,pyd+0.1);
							float Normal[2];
							Normal[0] = (Hgt_xp-Hgt_xm)/(0.2*GridSz_DEM);
							Normal[1] = (Hgt_yp-Hgt_ym)/(0.2*GridSz_DEM);
							float Nlen = std::sqrt(Normal[0]*Normal[0]+Normal[1]*Normal[1]+1);
							Normal[0] /= Nlen;
							Normal[1] /= Nlen;
							float cos_angle = Normal[0]*LightDir[0] + Normal[1]*LightDir[1] + LightDir[2];
							if (cos_angle < 0.0) cos_angle = 0.0;
							Correction = (1.0-cos_angle)-0.5;

							col[0] *= (1.0+Correction*LightAdjust[0]);
							col[1] *= (1.0+Correction*LightAdjust[1]);
							col[2] *= (1.0+Correction*LightAdjust[2]);
						}

						if (col[0] > 1.0) col[0] = 1.0;
						if (col[1] > 1.0) col[1] = 1.0;
						if (col[2] > 1.0) col[2] = 1.0;
						if (col[0] < 0.0) col[0] = 0.0;
						if (col[1] < 0.0) col[1] = 0.0;
						if (col[2] < 0.0) col[2] = 0.0;

						img_proc_s(px,py,0,0) = (unsigned char)(pow(col[0], 1.0/GAMMA)*255.0);
						img_proc_s(px,py,0,1) = (unsigned char)(pow(col[1], 1.0/GAMMA)*255.0);
						img_proc_s(px,py,0,2) = (unsigned char)(pow(col[2], 1.0/GAMMA)*255.0);
					}
				}
				redraw = true;
				break;
			}
			case cimg::keyARROWLEFT:
			{
				if (Adjust2)
					Exponent -= 0.1;
				else
					HMax -= 100;
				line_redraw = true;
				break;
			}
			case cimg::keyARROWRIGHT:
			{
				if (Adjust2)
					Exponent += 0.1;
				else
					HMax += 100;
				line_redraw = true;
				break;
			}
			case cimg::keyARROWUP:
			{
				if (Adjust < 3)
					Weights[Adjust] += 0.01;
				else
					LightAdjust[Adjust-3] += 0.1;
				line_redraw = true;
				break;
			}
			case cimg::keyARROWDOWN:
			{
				if (Adjust < 3)
					Weights[Adjust] -= 0.01;
				else
					LightAdjust[Adjust-3] -= 0.1;
				line_redraw = true;
				break;
			}
			case cimg::keyPAGEUP:
			{
				Brightness += 0.1;
				redraw = true;
				break;
			}
			case cimg::keyPAGEDOWN:
			{
				Brightness -= 0.1;
				redraw = true;
				break;
			}
			case cimg::keyE:
				Adjust2 = !Adjust2;
				line_redraw = true;
				break;
			case cimg::keyR:
				Adjust = 0;
				line_redraw = true;
				break;
			case cimg::keyG:
				Adjust = 1;
				line_redraw = true;
				break;
			case cimg::keyB:
				Adjust = 2;
				line_redraw = true;
				break;
			case cimg::keyT:
				Adjust = 3;
				line_redraw = true;
				break;
			case cimg::keyH:
				Adjust = 4;
				line_redraw = true;
				break;
			case cimg::keyN:
				Adjust = 5;
				line_redraw = true;
				break;
			case cimg::keyS:
			{
				Save = true;
			  stopflag = true;
				redraw = false;
				line_redraw = false;
				break;
			}
			case cimg::keyQ:
			case cimg::keyESC:
			{
				Save = false;
			  stopflag = true;
				redraw = false;
				line_redraw = false;
				break;
			}
		}

		if ((disp.mouse_x() != -1) && (disp.mouse_y() != -1))
		if ((disp.mouse_x() != Mx) && (disp.mouse_y() != My))
		{
			Mx = disp.mouse_x();
			My = disp.mouse_y();
			if (disp.mouse_x() > 0) line_redraw = true;
		}

		if (redraw)
		{
			if (disp_dem)
				img_disp = (img_dem_s.get_resize(dispsz_x, dispsz_y) - 32767)*((32767.0/9000.0)/255.0);
			else
				img_disp = img_proc_s*Brightness;

			line_redraw = true;
		}

		if (line_redraw)
		{
			int px = 0;
			int py = 0;
			int pxdem, pydem;
			int pxd, pyd;

			if (Mx >=0) px = Mx;
			if (My >=0) py = My;

			if (zoom)
			{
				pxd = Zx0 + px;
				pyd = Zy0 + py;
				pxdem = pxd*img_dem.width()/img_proc.width();
				pydem = pyd*img_dem.height()/img_proc.height();
			}
			else
			{
				pxd = px*img_col.width()/dispsz_x;
				pyd = py*img_col.height()/dispsz_y;
				pxdem = px*img_dem.width()/dispsz_x;
				pydem = py*img_dem.height()/dispsz_y;
			}

			float Height = img_dem(pxdem,pydem)/HMult-HZero;

			img_disp.draw_rectangle(0,0,img_disp.width(), 14, col_bkg,1.0f).
				draw_text(2,1,"Position (%d,%d)", col_text,col_bkg,1.0f,13,pxd, pyd).
				draw_text(120,1,"-> Height: %.1f  Color: <%d, %d, %d>", col_text1,col_bkg,1.0f,13,Height, img_col(pxd, pyd,0,0), img_col(pxd, pyd,0,1), img_col(pxd, pyd,0,2) ).
				draw_text(320,1,"Weights: <%.2f, %.2f, %.2f>", col_text3,col_bkg,1.0f,13,Weights[0], Weights[1], Weights[2]).
				draw_text(470,1,"Exponent: %.2f", col_text3,col_bkg,1.0f,13,Exponent).
				draw_text(560,1,"Ref. Height: %d", col_text3,col_bkg,1.0f,13,HMax).
				draw_text(660,1,"Light Fact: <%.2f, %.2f, %.2f>", col_text2,col_bkg,1.0f,13,LightAdjust[0], LightAdjust[1], LightAdjust[2]).
				draw_text(img_disp.width()-14,1,"[%d]", col_text1,col_bkg,1.0f,13,Adjust).
				draw_text(img_disp.width()-30,1,"[%s]", col_text1,col_bkg,1.0f,13,Adjust2?"E":"H").
				display(disp);
		}
		
	  if (disp.is_closed()) stopflag = true;
		redraw = false;
		line_redraw = false;
		disp.wait();
	}
	}

	if (Save)
	{
		float col[3];
		float Correction;

		bool doLight = true;
		if ((LightAdjust[0] < 0.01) && (LightAdjust[1] < 0.01) && (LightAdjust[2] < 0.01))
			doLight = false;

		std::fprintf(stderr,"Processing image...\n");
		cimg_forY(img_col,py)
		{
			float pyd = (float)py*img_dem.height()/img_col.height();
			cimg_forX(img_col,px)
			{
				float pxd = (float)px*img_dem.width()/img_col.width();

				float Height = img_dem.linear_atXY(pxd,pyd)/HMult-HZero;
				float AOT = ((float)HMax - Height)/float(HMax);
		
				col[0] = pow((float)img_col(px,py,0,0)/255.0, Gamma);
				col[1] = pow((float)img_col(px,py,0,1)/255.0, Gamma);
				col[2] = pow((float)img_col(px,py,0,2)/255.0, Gamma);

				col[0] -= pow(col[0], Exponent)*Weights[0]*AOT;
				col[1] -= pow(col[1], Exponent)*Weights[1]*AOT;
				col[2] -= pow(col[2], Exponent)*Weights[2]*AOT;

				if (doLight)
				{
					float Hgt_xm = img_dem.linear_atXY(pxd+0.1,pyd);
					float Hgt_xp = img_dem.linear_atXY(pxd-0.1,pyd);
					float Hgt_ym = img_dem.linear_atXY(pxd,pyd-0.1);
					float Hgt_yp = img_dem.linear_atXY(pxd,pyd+0.1);
					float Normal[2];
					Normal[0] = (Hgt_xp-Hgt_xm)/(0.2*GridSz_DEM);
					Normal[1] = (Hgt_yp-Hgt_ym)/(0.2*GridSz_DEM);
					float Nlen = std::sqrt(Normal[0]*Normal[0]+Normal[1]*Normal[1]+1);
					Normal[0] /= Nlen;
					Normal[1] /= Nlen;
					float cos_angle = Normal[0]*LightDir[0] + Normal[1]*LightDir[1] + LightDir[2];
					if (cos_angle < 0.0) cos_angle = 0.0;
					Correction = (1.0-cos_angle)-0.5;

					col[0] *= (1.0+Correction*LightAdjust[0]);
					col[1] *= (1.0+Correction*LightAdjust[1]);
					col[2] *= (1.0+Correction*LightAdjust[2]);
				}

				if (col[0] > 1.0) col[0] = 1.0;
				if (col[1] > 1.0) col[1] = 1.0;
				if (col[2] > 1.0) col[2] = 1.0;
				if (col[0] < 0.0) col[0] = 0.0;
				if (col[1] < 0.0) col[1] = 0.0;
				if (col[2] < 0.0) col[2] = 0.0;

				if (do_gamma)
				{
					img_proc(px,py,0,0) = (int)(pow(col[0], 1.0/Gamma)*(do_16?65535.0:255.0));
					img_proc(px,py,0,1) = (int)(pow(col[1], 1.0/Gamma)*(do_16?65535.0:255.0));
					img_proc(px,py,0,2) = (int)(pow(col[2], 1.0/Gamma)*(do_16?65535.0:255.0));		
				}
				else
				{
					img_proc(px,py,0,0) = (int)(col[0]*(do_16?65535.0:255.0));
					img_proc(px,py,0,1) = (int)(col[1]*(do_16?65535.0:255.0));
					img_proc(px,py,0,2) = (int)(col[2]*(do_16?65535.0:255.0));		
				}
			}
			std::fprintf(stderr,"\r  Line %d of %d (%d percent)...  ", py+1, img_col.height(), 100*(py+1)/img_col.height());
		}
		std::fprintf(stderr,"\n  Finished!\n");

		if (file_o != NULL)
		{
			std::fprintf(stderr,"Writing results...\n");
			img_proc.save(file_o);
			std::fprintf(stderr,"Results written to file %s\n", file_o);
		}
	}
}
